@@ -1,8 +1,8 @@
 #######################################################################
 #      $URL: svn+ssh://equilibrious@equilibrious.net/home/equilibrious/svnrepos/chrisdolan/Test-Virtual-Filesystem/lib/Test/Virtual/Filesystem.pm $
-#     $Date: 2007-11-14 00:40:26 -0600 (Wed, 14 Nov 2007) $
+#     $Date: 2007-11-16 01:29:27 -0600 (Fri, 16 Nov 2007) $
 #   $Author: equilibrious $
-# $Revision: 697 $
+# $Revision: 703 $
 ########################################################################
 
 package Test::Virtual::Filesystem;
@@ -19,7 +19,13 @@ use Attribute::Handlers;
 use Test::More;
 use base 'Test::Class';
 
-our $VERSION = '0.01';
+our $VERSION = '0.02';
+
+my $CAN_USE_XATTR;
+eval {
+   require File::ExtAttr;
+   $CAN_USE_XATTR = 1;
+};
 
 =pod
 
@@ -58,13 +64,26 @@ lots of typical filesystem operations.  This package attempts to
 accumulate a bunch of those tests to make it easier for you to test
 your filesystem.
 
+=head1 COMPATIBILITY POLICY
+
+Every time we add a new test to this suite, we annotate it with a
+version number.  If client code specifies an expected version number
+(say, 1.10) and it's running against a newer version or this module
+(say, 1.20) then any newer test will be marked as a TODO test.  That
+way if the test fails, it won't regress published code that used to
+work.
+
+This policy will allow us to continue adding new filesystem tests
+without worrying about breaking existing CPAN modules.
+
 =head1 CAVEATS AND LIMITATIONS
 
 This test class needs a more complete suite of test cases.  In
 particular, tests are needed for the following filesystem features:
 
     extended attributes (xattr)
-    symlinks
+    multiple symlinks
+    recursive symlinks
     hardlinks
     a/u/ctime
     nlink
@@ -74,7 +93,7 @@ particular, tests are needed for the following filesystem features:
     large files
     binary files
     files with awkward characters: EOF, NUL
-    non-ASCII filenames (maybe constructor should specify the encoding)
+    non-ASCII filenames (maybe constructor should specify the encoding?)
     permissions
     special file types (fifos, sockets, character and block devices, etc)
     file locking
@@ -87,16 +106,8 @@ particular, tests are needed for the following filesystem features:
     read, sysread, syswrite
     async I/O?
 
-=head1 COMPATIBILITY
-
-Until this package reaches v1.0, we will add arbitrary new tests which
-may break compatibility.  After 1.0, we will tag every test with a
-version number.  If client code specifies an expected version number
-(say, 1.10) and it's running against a newer version (say, 1.20) then
-any newer test that fails will be marked as a TODO test.
-
-This policy will allow us to continue adding new filesystem tests
-without worrying about breaking existing CPAN modules.
+Any help writing tests (or adapting tests from existing suites) will
+be appreciated!
 
 =head1 METHODS
 
@@ -123,9 +134,9 @@ This required property indicates where tests should run.
 
 Specify a Test::Virtual::Filesystem version number that is known to
 work.  If the actual Test::Virtual::Filesystem version number is
-greater, then any test cases added after the compatible version are
-considered C<TODO> tests.  See L<Test::More> for details about C<TODO>
-tests.
+greater, then any test cases added after the specified compatible
+version are considered C<TODO> tests.  See L<Test::More> for details
+about C<TODO> tests.
 
 =back
 
@@ -145,6 +156,10 @@ current value and return void.
 
 =over
 
+=item $self->test_all()
+
+Turns on all the tests.  Default false.
+
 =item $self->test_xattr()
 
 Default false.
@@ -201,26 +216,15 @@ Default false.
 
 =cut
 
-my $versions = {};
-sub Introduced : ATTR(CODE) { ##no critic(MixedCase)
-   my ($class, $symbol, $code_ref, $attr, $args) = @_;
-   if ($symbol eq 'ANON') {
-      warn 'cannot test anonymous subs - you probably loaded ' . __PACKAGE__ . ' too late.' .
-          ' (after the CHECK block was run)';
-   } else {
-      my $name = *{$symbol}{NAME};
-      $versions->{$class} ||= {};
-      $versions->{$class}->{$name} = $args;
-   }
-   return;
-}
-
 sub new {
    my ($pkg, $opts) = @_;
    my $self = $pkg->SUPER::new();
    $opts ||= {};
-   $self->{mountdir} = $opts->{mountdir};
+   for my $key (qw(mountdir compatible)) {
+      $self->{$key} = $opts->{$key};
+   }
    $self->{fs_opts} = {
+      'all' => 0,
       'xattr' => 0,
       'time' => {
          'atime' => 0,
@@ -268,6 +272,12 @@ sub init {
    }
 }
 
+=item setup()
+
+Runs before every test to prepare a directory for testing.
+
+=cut
+
 sub setup : Test(setup => 1) {
    my ($self) = @_;
    if (!defined $self->{mountdir}) {
@@ -284,18 +294,22 @@ sub setup : Test(setup => 1) {
           '  I am too scared of deleting important files to use it';
    }
    $self->{tempdir} = File::Spec->catdir($self->{mountdir}, 'testdir');
-       #File::Temp::tempdir('testfs_XXXX', DIR => $self->{mountdir}); # may croak!
    mkdir $self->{tempdir};
    ok(-d $self->{tempdir}, 'Created tempdir');
    return;
 }
 
+=item teardown()
+
+Runs after every test to clean up the test directory so the next test
+will have a clean workspace.
+
+=cut
+
 sub teardown : Test(teardown => 1) {
    my ($self) = @_;
-   #system('find', $self->{mountdir});
    my $tmpdir = delete $self->{tempdir};
    if (defined $tmpdir && -e $tmpdir) {
-      #File::Path::rmtree($tmpdir);
       $self->_cleandir($tmpdir);
       if ($tmpdir ne $self->{mountdir}) {
          rmdir $tmpdir or die $OS_ERROR;
@@ -323,20 +337,146 @@ sub _cleandir {
    return;
 }
 
+=item Introduced($version)
+
+A subroutine attribute used to flag the Test::Virtual::Filesystem
+version number when that test was introduced.  It's used like this:
+
+  sub open_nonexistent_file : Tests(1) : Introduced('0.02') {
+     ok(!open(my $f, '<', '/tmp/no_such_file'));
+  }
+
+=cut
+
+sub Introduced : ATTR(CODE) { ##no critic(MixedCase)
+   my ($class, $symbol, $code_ref, $attr, $introduced_version) = @_;
+   if ($symbol eq 'ANON') {
+      warn 'cannot test anonymous subs - you probably loaded ' . __PACKAGE__ . ' too late.' .
+          ' (after the CHECK block was run)';
+   } else {
+      # Wrap the sub in a version test
+      no warnings 'redefine';  ##no critic(TestingAndDebugging::ProhibitNoWarnings)
+      *{$symbol} = sub {
+         local $TODO = $_[0]->_compatible($introduced_version); ## no critic(Local)
+         $code_ref->(@_);
+      };
+      #print STDERR "record $class\::$name as $args\n";
+   }
+   return;
+}
+sub _compatible {
+   my ($self, $introduced_version) = @_;
+   return if !$self->{compatible};
+   return if $introduced_version le $self->{compatible};
+   return 'compatibility mode ' . $self->{compatible};
+}
+
+=item Features($featurelist)
+
+This is a subroutine attribute to specify one or more features used in
+the test.  The features should be listed as a comma-separated list:
+
+  sub symlink_create : Tests(1) : Features('symlink') {
+     ok(symlink($src, $dest));
+  }
+  sub symlink_permissions : Tests(2) : Features('symlink, permissions') {
+     ok(symlink($src, $dest));
+     ok(-w $dest);
+  }
+
+Subfeatures must be separated from their parent features by a C</>.  For example:
+
+  sub atime_utime_set : Tests(1) : Features('time/atime, time/utime') {
+     my $now = time;
+     ok(utime($now, $now, $file));
+  }
+
+Look at the source code for C<new()> to see the supported features and
+subfeatures.  The C<test_*> methods above describe the all the
+features, but in those methods the subfeature names are flattened.
+
+=cut
+
+sub Features : ATTR(CODE) { ##no critic(MixedCase)
+   my ($class, $symbol, $code_ref, $attr, $features) = @_;
+   if ($symbol eq 'ANON') {
+      warn 'cannot test anonymous subs - you probably loaded ' . __PACKAGE__ . ' too late.' .
+          ' (after the CHECK block was run)';
+   } else {
+      my @features = split m/,\s*/xms, $features;
+      my $name = *{$symbol}{NAME};
+      # Wrap the sub in a feature test
+      no warnings 'redefine';  ##no critic(TestingAndDebugging::ProhibitNoWarnings)
+      *{$symbol} = sub {
+       SKIP: {
+          my $blocking_feature = $_[0]->_blocking_feature(@features);
+          if ($blocking_feature) {
+             my $numtests = __PACKAGE__->num_method_tests($name);
+             skip 'feature unsupported: ' . $blocking_feature, $numtests;
+          }
+          $code_ref->(@_);
+         }
+      };
+   }
+   return;
+}
+sub _blocking_feature {
+   my ($self, @features) = @_;
+   return if $self->{fs_opts}->{all};
+   for my $feature (@features) {
+      my $opts = $self->{fs_opts};
+      for my $part (split m{/}xms, $feature) {
+         return $feature if !ref $opts;
+         return $feature if !$opts->{$part};
+         $opts = $opts->{$part};
+      }
+   }
+   return;
+}
+
+
+sub _compatibility_test : Test(1) : Introduced('0.02') {
+   # introduced in v0.02 I<solely> to test the compatibility feature.
+   my ($self) = @_;
+   pass('compatibility_test');
+   return;
+}
+
+sub _features_test : Test(1) : Introduced('0.02') : Features('features_test') {
+   # introduced in v0.02 I<solely> to test the 'features' feature.
+   my ($self) = @_;
+   pass('features_test');
+   return;
+}
+
+
 =item stat_dir(), introduced in v0.01
 
 =cut
 
-sub stat_dir : Test(4) : Introduced('0.01') {
+sub stat_dir : Test(6) : Introduced('0.01') {
    my ($self) = @_;
    my $f = $self->_file(q{/});
    ok(-e $f, 'mount dir exists');
    ok(-d $f, 'mount dir is a dir');
+   ok(!-f $f, 'mount dir is not a file');
+   ok(!-l $f, 'mount dir is not a symlink');
    ok(-r $f, 'mount dir is readable');
    ok(-x $f, 'mount dir is searchable');
    return;
 }
 
+=item stat_dir_size(), introduced in v0.02
+
+=cut
+
+sub stat_dir_size : Test(1) : Introduced('0.02') {
+   my ($self) = @_;
+   my $f = $self->_file(q{/});
+   ok(-s $f, 'mount dir has non-zero size');
+   return;
+}
+
 =item read_dir(), introduced in v0.01
 
 =cut
@@ -440,6 +580,8 @@ sub write_append_file : Test(2) : Introduced('0.01') {
    return;
 }
 
+## Perl's '>>' creates the file if it doesn't exist
+#
 #=item write_append_file_fail(), introduced in v0.01
 #
 #=cut
@@ -473,12 +615,13 @@ sub write_read_file : Test(1) : Introduced('0.01') {
 
 =cut
 
-sub write_unlink_file : Test(2) : Introduced('0.01') {
+sub write_unlink_file : Test(3) : Introduced('0.01') {
    my ($self) = @_;
    my $f = $self->_file('/read_file');
    my $content = "content\n";
    $self->_write_file($f, $content);
-   ok(-f $f, 'file exists');
+   ok(-e $f, 'file exists');
+   ok(-f $f, 'file is a file');
    unlink $f or die $OS_ERROR;
    ok(!-f $f, 'file is deleted');
    return;
@@ -542,6 +685,89 @@ sub write_subdir : Test(3) : Introduced('0.01') {
    return;
 }
 
+=item symlink_create(), introduced in v0.02
+
+=cut
+
+sub symlink_create : Test(10) : Introduced('0.02') : Features('symlink') {
+   my ($self) = @_;
+   my $src = $self->_file('/symlink_target');
+   mkdir $src or die $OS_ERROR;
+   ok(-e $src, 'symlink source exists');
+   my $s = $self->_file('/symlink_create');
+   ok(symlink($src, $s), 'created symlink') or die $OS_ERROR;
+   ok(-e $s, 'symlink exists');
+   ok(-l $s, 'symlink is a symlink');
+   ok(-d $s, 'symlink src is a dir');
+   ok(!-f $s, 'symlink src is not a file');
+   is(-s $s, -s $src, 'symlink size is size of source');
+   is(readlink $s, $src, 'read newly created symlink');
+   unlink $s or die $OS_ERROR;
+   ok(!-e $s, 'symlink deleted');
+   ok(-e $src, 'symlink source not deleted');
+   return;
+}
+
+=item xattr_list(), introduced in v0.02
+
+=cut
+
+sub xattr_list : Test(1) : Introduced('0.02') : Features('xattr') {
+   my ($self) = @_;
+ SKIP: {
+      if (!$CAN_USE_XATTR) {
+         skip 'Optional File::ExtAttr not found', $self->num_method_tests('xattr_list');
+      }
+      my $f = $self->_file(q{/});
+      my @attrs = File::ExtAttr::listfattr($f);
+      ok(@attrs == 0 || defined $attrs[0], 'got xattr list');
+   }
+   return;
+}
+
+=item xattr_set(), introduced in v0.02
+
+=cut
+
+sub xattr_set : Test(9) : Introduced('0.02') : Features('xattr') {
+   my ($self) = @_;
+ SKIP: {
+      if (!$CAN_USE_XATTR) {
+         skip 'Optional File::ExtAttr not found', $self->num_method_tests('xattr_list');
+      }
+      my $f = $self->_file('/foo');
+      $self->_write_file($f);
+
+      my $xattr_key = 'org.cpan.cdolan';
+      my $xattr_value = 'test';
+      my $xattr_replace = 'test2';
+
+      # just in case, clean up!
+      if (defined File::ExtAttr::getfattr($f, $xattr_key)) {
+         File::ExtAttr::delfattr($f, $xattr_key);
+      }
+      {
+         # File::ExtAttr doesn't look at $^W or 'no warnings'.  Grr...
+         local $SIG{__WARN__} = sub {};
+         ok(!File::ExtAttr::setfattr($f, $xattr_key, $xattr_value, {replace => 1}), 'cannot replace missing xattr');
+      }
+      ok(File::ExtAttr::setfattr($f, $xattr_key, $xattr_value, {create => 1}), 'create xattr');
+      is(File::ExtAttr::getfattr($f, $xattr_key), $xattr_value, 'get xattr');
+      ok((any {$xattr_key eq $_} File::ExtAttr::listfattr($f)), 'list xattr');
+      {
+         local $SIG{__WARN__} = sub {};
+         ok(!File::ExtAttr::setfattr($f, $xattr_key, $xattr_value, {create => 1}), 'cannot create existing xattr');
+      }
+      ok(File::ExtAttr::setfattr($f, $xattr_key, $xattr_replace, {replace => 1}), 'replace xattr');
+      is(File::ExtAttr::getfattr($f, $xattr_key), $xattr_replace, 'get xattr');
+      ok(File::ExtAttr::delfattr($f, $xattr_key), 'delete xattr');
+      is(File::ExtAttr::getfattr($f, $xattr_key), undef, 'xattr deleted');
+
+      unlink $f;
+   }
+   return;
+}
+
 #       xattr => 1,
 #       time => {
 #          utime => 1,
@@ -622,7 +848,8 @@ Chris Dolan, I<cdolan@cpan.org>
 =cut
 
 # Local Variables:
-#   mode: cperl
+#   mode: perl
+#   perl-indent-level: 3
 #   cperl-indent-level: 3
 #   fill-column: 78
 #   indent-tabs-mode: nil
